# -*- coding: UTF-8 -*-
import lief
import os


__all__ = ["REMINDer"]


THRESHOLDS = {
    'default': 6.85,
    'PE':      6.85,
}


class REMINDer:
    def __init__(self, entropy_threshold=None, logger=None, **kwargs):
        """ Configure the detector with various parameters. """
        self.__entropy = entropy_threshold
        self.logger = logger
    
    def detect(self, executable):
        """ Analyze the input executable file using the custom heuristic. """
        # parse the input executable using LIEF (catch warnings from stderr)
        tmp_fd, null_fd = os.dup(2), os.open(os.devnull, os.O_RDWR)
        os.dup2(null_fd, 2)
        binary = lief.parse(str(executable))
        os.dup2(tmp_fd, 2)  # restore stderr
        os.close(null_fd)
        if binary is None:
            raise TypeError("Not an executable")
        etype = "ELF" if type(binary) is lief.ELF.Binary else \
                "Mach-O" if type(binary) is lief.MachO.Binary else \
                "PE" if type(binary) is lief.PE.Binary else None
        # locate the entry point and its section
        ep = binary.rva_to_offset(binary.optional_header.addressof_entrypoint)
        ep_section = binary.section_from_rva(binary.optional_header.addressof_entrypoint)
        section_name = ep_section.name.rstrip("\x00")
        if self.logger:
            self.logger.debug("EP at 0x%.8x in %s" % (ep, section_name))
        # now apply the heuristic from https://ieeexplore.ieee.org/document/5404211
        # 1) check if the EP section is writable
        if ep_section.has_characteristic(lief.PE.SECTION_CHARACTERISTICS.MEM_WRITE):
            threshold = self.__entropy or THRESHOLDS.get(etype, THRESHOLDS['default'])
            if self.logger:
                self.logger.debug("%s is writable" % section_name)
                self.logger.debug("entropy = %.3f (threshold: %.3f)" % (ep_section.entropy, threshold))
            # 2) check if the entropy of the EP section is above a given threshold
            if ep_section.entropy > threshold:
                return True
        return False




# -*- coding: UTF-8 -*-
import logging
from argparse import ArgumentParser, RawTextHelpFormatter
from ast import literal_eval
from os.path import exists
from time import perf_counter

from .__info__ import __author__, __copyright__, __license__, __reference__, __source__, __version__
from .__init__ import REMINDer


def valid_file(path):
    if not exists(path):
        raise ValueError("input file does not exist")
    return path


class Positive:
    def __init__(self, *types):
        self._types = types
    
    def __call__(self, string):
        try:
            n = literal_eval(string)
        except ValueError:
            raise ValueError(string)
        if not isinstance(n, self._types) or n < 0.:
            raise ValueError(string)
        return self._types[0](n)
    
    def __repr__(self):
        return "positive %s" % "|".join(map(lambda x: x.__name__, self._types))


def main():
    """ Tool's main function """
    descr = "REMINDer {}\n\nAuthor   : {}\nCopyright: {}\nLicense  : {}\nReference: {}\nSource   : {}\n" \
            "\nThis tool applies a custom heuristic based on the entropy of the Entry Point section to determine if" \
            " an executable is packed or not.\n\n"
    descr = descr.format(__version__, __author__, __copyright__, __license__, __reference__, __source__)
    examples = "usage examples:\n- " + "\n- ".join([
        "reminder program.exe",
        "reminder /bin/ls --entropy-threshold 6.9",
    ])
    parser = ArgumentParser(description=descr, epilog=examples, formatter_class=RawTextHelpFormatter, add_help=False)
    parser.add_argument("path", type=valid_file, help="path to the executable file")
    opt = parser.add_argument_group("optional arguments")
    opt.add_argument("--entropy-threshold", type=Positive(float, int),
                     help="threshold for the entropy of the Entry Point section")
    extra = parser.add_argument_group("extra arguments")
    extra.add_argument("-b", "--benchmark", action="store_true",
                       help="enable benchmarking, output in seconds (default: False)")
    extra.add_argument("-h", "--help", action="help", help="show this help message and exit")
    extra.add_argument("-v", "--verbose", action="store_true", help="display debug information (default: False)")
    args = parser.parse_args()
    logging.basicConfig()
    args.logger = logging.getLogger("reminder")
    args.logger.setLevel([logging.INFO, logging.DEBUG][args.verbose])
    code = 0
    # execute the tool
    if args.benchmark:
        t1 = perf_counter()
    try:
        r = REMINDer(**vars(args)).detect(args.path)
        dt = str(perf_counter() - t1) if args.benchmark else ""
        print(str(r))
        if dt != "":
            print(dt)
    except Exception as e:
        if "magic not found." in str(e):
            e.value = "Not a valid executable file"
        if str(e) != "no result":
            args.logger.exception(e)
        code = 1
    return code




# -*- coding: UTF-8 -*-
"""REMINDer package information.

"""
import os
from datetime import datetime

__y = str(datetime.now().year)
__s = "2022"

__author__    = "Alexandre D'Hondt"
__copyright__ = "© {} Alexandre D'Hondt".format([__y, __s + "-" + __y][__y != __s])
__license__   = "GPLv3 (https://www.gnu.org/licenses/gpl-3.0.fr.html)"
__reference__ = "https://ieeexplore.ieee.org/document/5404211"
__source__    = "https://github.com/packing-box/REMINDer"

with open(os.path.join(os.path.dirname(__file__), "VERSION.txt")) as f:
    __version__ = f.read().strip()

